#include "robotiq_2f_controllers/robotiq_gripper_command_controller.hpp"


using config_type = controller_interface::interface_configuration_type;

namespace robotiq_2f_controllers
{
RobotiqForwardController::RobotiqGripperCommandController() : controller_interface::ControllerInterface() {}

controller_interface::CallbackReturn RobotiqGripperCommandController::on_init()
{
    joint_name_ = auto_declare<std::string>("joint", joint_name_);
    command_interface_types_ =
    auto_declare<std::vector<std::string>>("command_interfaces", command_interface_types_);
    state_interface_types_ =
    auto_declare<std::vector<std::string>>("state_interfaces", state_interface_types_);


    if (joint_name_.empty()) {
        RCLCPP_ERROR(get_node()->get_logger(), "Missing 'joint_name' parameter.");
        return CallbackReturn::ERROR;
    }
    else{
        RCLCPP_INFO(get_node()->get_logger(), "Joint name: %s.", joint_name_.c_str());
    }

    return CallbackReturn::SUCCESS;
}

controller_interface::InterfaceConfiguration RobotiqGripperCommandController::command_interface_configuration() const
{
    RCLCPP_DEBUG(get_node()->get_logger(), "Starting configuration of command interfaces.");

    controller_interface::InterfaceConfiguration config;
    config.type = controller_interface::interface_configuration_type::INDIVIDUAL;

    for (const auto& interface_map_entry : command_interface_map_) {
        const std::string& interface_type = interface_map_entry.first;
        config.names.push_back(joint_name_ + "/" + interface_type);
    }
    RCLCPP_DEBUG(get_node()->get_logger(), "Completed configuration of command interfaces.");
    return config;
}


controller_interface::InterfaceConfiguration RobotiqGripperCommandController::state_interface_configuration() const
{
    RCLCPP_DEBUG(get_node()->get_logger(), "Starting configuration of state interfaces.");

    controller_interface::InterfaceConfiguration config;
    config.type = controller_interface::interface_configuration_type::INDIVIDUAL;

    for (const auto& interface_map_entry : state_interface_map_) {
        const std::string& interface_type = interface_map_entry.first;
        config.names.push_back(joint_name_ + "/" + interface_type);
    }

    RCLCPP_DEBUG(get_node()->get_logger(), "Completed configuration of state interfaces.");
    return config;
}

controller_interface::CallbackReturn RobotiqGripperCommandController::on_configure(const rclcpp_lifecycle::State &)
{
    this->action_server_ = rclcpp_action::create_server<robotiq_2f_msgs::action::GripperCommand>(
        get_node(),
        "~/gripper_command",
        std::bind(&RobotiqGripperCommandController::handle_goal, this, std::placeholders::_1, std::placeholders::_2),
        std::bind(&RobotiqGripperCommandController::handle_cancel, this, std::placeholders::_1),
        std::bind(&RobotiqGripperCommandController::handle_accepted, this, std::placeholders::_1)
    );
    return CallbackReturn::SUCCESS;
}

controller_interface::CallbackReturn RobotiqGripperCommandController::on_activate(const rclcpp_lifecycle::State &)
{
    RCLCPP_INFO(get_node()->get_logger(), "Available command interfaces:");
    for (const auto& interface : command_interfaces_) {
        RCLCPP_INFO(get_node()->get_logger(), "Interface Name: %s, Interface Type: %s", interface.get_name().c_str(), interface.get_interface_name().c_str());
    }

    // clear out vectors in case of restart
    joint_position_command_interface_.clear();
    max_velocity_command_interface_.clear();
    max_effort_command_interface_.clear();

    // assign command interfaces
    for (auto & interface : command_interfaces_)
    {
        command_interface_map_[interface.get_interface_name()]->push_back(interface);
    }

    return CallbackReturn::SUCCESS;
}

controller_interface::return_type RobotiqGripperCommandController::update(const rclcpp::Time& /*time*/, const rclcpp::Duration& /*period*/)
{
    /*
    if (new_msg_)
    {
        command_msg_ = *msg_external_ptr_.readFromRT();
        desired_position_ = command_msg_->position;
        desired_max_velocity_ = command_msg_->max_velocity;
        desired_max_effort_ = command_msg_->max_effort;
    }
    // Set the hardware interface commands
    if (joint_position_command_interface_.empty()) {
        RCLCPP_FATAL(get_node()->get_logger(), "Command interface is not available.");
        return controller_interface::return_type::ERROR;
    }

    joint_position_command_interface_[0].get().set_value(desired_position_);
    max_velocity_command_interface_[0].get().set_value(desired_max_velocity_);
    max_effort_command_interface_[0].get().set_value(desired_max_effort_);

    return controller_interface::return_type::OK;
    */
}


controller_interface::CallbackReturn RobotiqGripperCommandController::on_deactivate(const rclcpp_lifecycle::State &)
{
  release_interfaces();

  return CallbackReturn::SUCCESS;
}

controller_interface::CallbackReturn RobotiqGripperCommandController::on_cleanup(const rclcpp_lifecycle::State &)
{
  return CallbackReturn::SUCCESS;
}

controller_interface::CallbackReturn RobotiqGripperCommandController::on_error(const rclcpp_lifecycle::State &)
{
  return CallbackReturn::SUCCESS;
}

controller_interface::CallbackReturn RobotiqGripperCommandController::on_shutdown(const rclcpp_lifecycle::State &)
{
  return CallbackReturn::SUCCESS;
}
}

#include "pluginlib/class_list_macros.hpp"
PLUGINLIB_EXPORT_CLASS(robotiq_2f_controllers::RobotiqGripperCommandController, controller_interface::ControllerInterface)

